@@ -220,6 +220,46 @@ static void test_5p1_N(void)
 			       0.0, 0.0, 0.0, 0.0, 0.0, 1.0));
 }
 
+static void test_6p1_N(void)
+{
+	test_mix(7, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(RC)|_M(SL)|_M(SR), 1, _M(MONO), 0,
+			MATRIX(0.707107, 0.707107, 1.0, 0.0, 0.5, 0.5, 0.5));
+	test_mix(7, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(SL)|_M(SR)|_M(RC),
+		 6, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(SL)|_M(SR), 0,
+			MATRIX(1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.707107,
+			       0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.707107));
+	test_mix(7, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(SL)|_M(SR)|_M(RC),
+		 6, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(RL)|_M(RR), 0,
+			MATRIX(1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.707107,
+			       0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.707107));
+	test_mix(7, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(RC)|_M(RL)|_M(RR),
+		 6, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(RL)|_M(RR), 0,
+			MATRIX(1.0, 0.0, 0.0, 0.0, 0.0,      0.0, 0.0,
+			       0.0, 1.0, 0.0, 0.0, 0.0,      0.0, 0.0,
+			       0.0, 0.0, 1.0, 0.0, 0.0,      0.0, 0.0,
+			       0.0, 0.0, 0.0, 1.0, 0.0,      0.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 0.707107, 1.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 0.707107, 0.0, 1.0));
+	test_mix(7, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(SL)|_M(SR)|_M(RC),
+		 8, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(SL)|_M(SR)|_M(RL)|_M(RR), 0,
+			MATRIX(1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0,
+			       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.707107,
+			       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.707107));
+}
+
 static void test_7p1_N(void)
 {
 	test_mix(8, _M(FL)|_M(FR)|_M(LFE)|_M(FC)|_M(SL)|_M(SR)|_M(RL)|_M(RR), 1, _M(MONO), 0,
@@ -316,6 +356,7 @@ int main(int argc, char *argv[])
 	test_3p1_N();
 	test_4_N();
 	test_5p1_N();
+	test_6p1_N();
 	test_7p1_N();
 
 	test_n_m_impl();
